<?php

require_once(dirname(__DIR__) . "/includes/base.inc.php");
//require_once(LIBS . '/Mustache/init.php');
require_once(INCLUDES . '/responses.inc.php');


/*
 * View Exceptions
 */
class ViewError extends BaseError {
    public $name = 'View Error';
    public $code = ErrorCodes::VIEW_ERROR;
}
class MissingData extends ViewError {
    public $name = 'Missing Required Data';
}
class InvalidData extends ViewError {
    public $name = 'Invalid Data';
}
class InvalidRoute extends ViewError {
    public $name = 'Invalid Route';
}
class InvalidMethod extends ViewError {
    public $name = 'Invalid Method';
}
class InvalidAction extends ViewError {
    public $name = 'Not Allowed';
}
class RouteDeprecated extends ViewError {
    public $name = 'Route Deprecated';
}
class RouteNotImplemented extends ViewError {
    public $name = 'Route Not Implemented';
}


class View {
    public $dbg;

    public $page = null;
    public $route = null; // Stores $action for POST requests, $page for GET
    public $endpoint = null; // Stores the final endpoint destination
    public $method = null; // Request Method
    public $token = null; // Token used for logging in.
    public $action = null; // Stores action to be executed at final endpoint/route
    public $tracking = null; // Store the tracking number for ease of finding issues
    public $notes = null; // Notes that can be added by the caller

    // GET and POST are stored here for easy access
    public $DATA = null;

    // Used to define the unique ID field for this View
    protected $uID = null;
    // ID is removed from incoming request data and placed into $View->ID for easy access
    public $ID = null;

    // Set default page and define the map for GET requests
    protected $defaultPage = '';
    protected $pageMap = array();

    // Define the map for POST requests
    protected $routeMap = array();

    // Set Authentication for View
    protected $auth = null;

    // Any stylesheet or script paths placed in these arrays will be built and included in the header
    protected $CSS = array();
    protected $JS = array();

    protected $header = '/header.php';
    protected $footer = '/footer.php';

    function __construct($dbg = false) {
        $this->dbg = $dbg;
        // Wrap errorHandler around anything that can throw errors
        $this->errorHandler(function () {
            // Check Request Method
            $this->checkMethod($_SERVER['REQUEST_METHOD']);

            // Send a JSON response if the client requests it
            if ($this->acceptsMime('application/json')){
                header('Content-type: application/json');
            }

            header('Access-Control-Allow-Origin: *');
            header("Access-Control-Allow-Methods: GET, POST, PUT, DELETE, OPTIONS");
            header("Access-Control-Allow-Headers: Origin, X-Requested-With, Content-Type, Accept, Authorization");


            //Allows for cross domain requests in CORS
            header('Access-Control-Allow-Origin: *');

            // Grab request data
            $this->readRequestData();

            // Check authorization at the View level
            if ($this->auth)
                $this->checkAuth($this->auth);

            // Run initialization function
            $this->init();

            // Handle request
            if ($this->method == 'GET')
                return $this->serve($this->route);
            else if ($this->method == 'POST')
                return $this->routeRequest($this->route);
            else
                throw new InvalidMethod("The {$this->method} method is not supported by the SG API!");
        });
    }

    public function checkMethod($method) {
        $this->method = $method;
    }

    public function readRequestData() {
        if ($this->method == 'GET') {
            $this->route = (isset($_GET['page']) ? $_GET['page'] : $this->defaultPage);
            $this->DATA = $_GET;
        }
        else if ($this->method == 'POST') {
            if ($this->contentType('application/json'))
                $requestData = json_decode(file_get_contents("php://input"), true);
            else
                $requestData = $_POST;

            $this->route = $requestData['route'];
            $this->DATA = $requestData['data'];
            $this->token = $requestData['token'];
            $this->endpoint = $requestData['endpoint'];
            $this->action = $requestData['action'];
            $this->tracking = $requestData['tracking'];
            $this->notes = $requestData['notes'];

            $logID = LogRequest($this->tracking, $this->route, $this->endpoint, $this->action, $this->notes);
            $_SESSION['log_id'] = $logID;
        }
    }

    public function serve($route) {
        // Serves an HTML response to the client, complete with header, resources and footer

        // Prep resources for insertion into header
        extract(self::prepResources($this->CSS, $this->JS));

        // Use output buffering to suppress html until we have generated the entire page
        ob_start();

        // Begin filling the buffer
        if ($this->header)
            require_once(TEMPLATES . $this->header);
        echo $this->renderRoute($route);
        if ($this->footer)
            require_once(TEMPLATES . $this->footer);

        // Grab the buffer and destroy it
        $html = ob_get_contents();
        ob_end_clean();

        return $html;
    }

    /**
     * @param array $CSS
     * @param array $JS
     * @return array
     *
     * Returns
     */
    public static function prepResources($CSS = array(), $JS = array()) {
        // Wrap resource paths in their respective tags
        return array(
            'CSS' => "<link href='" . implode("' rel='stylesheet' type='text/css' /><link href='", $CSS) . "' rel='stylesheet' type='text/css' />",
            'JS' => "<script src='" . implode("' type='text/javascript'></script><script src='", $JS) . "' type='text/javascript'></script>"
        );
    }

    /**
     * This function is called before the incoming request is processed, but after the View authentication check.
     * Usage: Override this stub in children classes to run code any time the View is initialized
     */
    public function init() {
        return;
    }

    public function routeRequest($route) {
        if (!isset($route))
            $route = "defaultRoute";

        if (isset($this->routeMap[$route])) {
            $action = $this->routeMap[$route];

            // Check authorization at the Route level
            if (isset($action['auth']))
                $this->checkAuth($action['auth']);

            // Make sure required data is present
            if (isset($action['req']))
                $this->checkReq($action['req']);

            // Execute route function
            return $this->$action['fn']();
        }
        else
            throw new InvalidRoute("The route [{$route}] does not exist at this endpoint!");
    }

    public function renderRoute($route) {
        // Build HTML using the template and templateData for this View
        // Templates located in /templates directory
        if (isset($this->pageMap[$route])) {
            $page = $this->pageMap[$route];

            // Check authorization at the Page level
            if (isset($page['auth']))
                $this->checkAuth($page['auth']);

            // Check for templateData function
            $data = (isset($page['fn']) ? $this->$page['fn']() : array());

            // Execute route and return response
            return $this->renderTemplate($page['template'], $data);
        }
        else
            throw new InvalidRoute("The page [{$route}] does not exist at this endpoint!");
    }

    public function renderTemplate($templateFile, $data) {
        global $Mustache;
        return $Mustache->render($templateFile, $data);
    }

    protected function checkAuth($auth) {
        // Ensure the current user is authenticated as $auth
        // Throw an error if user doesn't meet authentication requirement
        // $auth = 'user' || $auth = array('admin', 'ocpos')

        // Allow client to send auth token in request
        if (isset($this->token) && $this->token != '')
            session_id($this->token);

        if (is_array($auth)) {
            foreach ($auth as $a)
                $this->checkAuth($a);
        }

        if($auth == 'admin' && $_SESSION['locationID'] != 1){
            throw new AuthDenied('You must be authenticated as the admin user to access this route!');
        }
        else {
            if ($auth == 'location' && !AuthAPI::isLocation())
                //return json_encode(session_id());
                throw new AuthDenied('Must be authenticated before proceeding.');
        }
    }

    protected function checkReq($req) {
        // Ensure all values in $req are present in $_POST
        // Throw an error if any data is missing or empty
        // $req = 'userID' || $req = array('userID', 'locID', 'someData' => array('type' => 'int')

        if (is_array($req))
            foreach ($req as $k => $v) {
                if (!is_int($k)) // Is this array numeric?
                    // Type is specified
                    $this->requireData($k, $v);
                else
                    // No type specified
                    $this->requireData($v);
            }
        else $this->requireData($req);
        return True;
    }

    private function requireData($req, $type = null) {
        if (!isset($this->DATA[$req]) || $this->DATA[$req] == '')
            // If not set or empty
            throw new MissingData("[{$req}] is required");
        if ($type !== null)
            $this->checkType($req, $type);
        if ($this->uID == $req) {
            // Detect and grab uID for this View
            $this->ID = $this->DATA[$req];
            unset($this->DATA[$req]);
        }
    }

    private function checkType($name, $type) {
        $val = $this->DATA[$name];
        if ($type == 'bool' && !$this->isBool($val))
            throw new InvalidData("[{$name}] must be a Boolean");
        else if ($type == 'num' && !is_numeric($val))
            throw new InvalidData("[{$name}] must be Numeric");
        else if ($type == 'int' && !$this->isInt($val))
            throw new InvalidData("[{$name}] must be an Integer");
        else if ($type == 'float' && !$this->isFloat($val))
            throw new InvalidData("[{$name}] must be a Float");
        else if ($type == 'str' && (is_numeric($val) || is_array($val)))
            throw new InvalidData("[{$name}] must be a String");
    }

    private function isInt($i) {
        if (is_int($i))
            return true;
        elseif (is_string($i))
            return ctype_digit($i);
        else
            return false;
    }

    private function isFloat($f) {
        return !$this->isInt($f) && filter_var($f, FILTER_VALIDATE_FLOAT);
    }

    private function isBool($b) {
        switch (strtolower($b)) {
            case 'true':
            case '1':
            case 'false':
            case '0':
                return true;
            default:
                return false;
        }
    }

    /**
     * @param $op
     * @param null $type
     * @return bool
     *
     * Returns the value of $op if sent by the client, otherwise return false
     */
    protected function option($op, $type = null) {
        if (isset($this->DATA[$op]) && $this->DATA[$op] !== '') {
            if ($type !== null)
                $this->checkType($op, $type);
            return $this->DATA[$op];
        }
        else
            return false;
    }

    /**
     * @param $mimeType
     * @return bool
     *
     * Returns true if $mimeType exists in the [Accept] header sent by the client
     */
    private function acceptsMime($mimeType) {
        // Returns true if $mimeType is present in the Accept header of the client request
        $accepted = explode(',', (strtolower(str_replace(' ', '', $_SERVER['HTTP_ACCEPT']))));
        if ($accepted)
            foreach ($accepted as $mime)
                if ($mime == $mimeType)
                    return true;
        return false;
    }

    /**
     * @param $contentType
     * @return bool
     *
     * Returns true if $contentType exists in the [Content-Type] header sent by the client
     */
    private function contentType($contentType) {
        // Returns true if $contentType is set as the Content-type of the request
        $content = explode(';', (strtolower(str_replace(' ', '', $_SERVER['CONTENT_TYPE']))));
        if ($content)
            foreach ($content as $mime)
                if ($mime == $contentType)
                    return true;
        return false;
    }

    /**
     * @param $data
     * @return array
     *
     * Prepares the given data to be returned to the client
     */
    public static function prepResponseData($data) {
        $prepped = array();
        if ($data !== null) {
            // If array, prep all elements of array
            if (is_array($data)) {
                foreach ($data as $k => $v)
                    $prepped[$k] = static::prepResponseData($v);
            }
            else {
                // If Model, export model data
                if ($data instanceof DBModel)
                    $prepped = $data->export();
                else
                    $prepped = $data;
            }
        }

        return $prepped;
    }

    /**
     * @param int $keyLength
     * @return string
     *
     * Returns a token string which can be used to generate one-time links
     * Warning: These tokens do not take any security measures into account
     */
    public static function genToken($keyLength = 32) {
        $chars = "abcdefghijklmnopqrstuwxyzABCDEFGHIJKLMNOPQRSTUWXYZ0123456789";
        $pass = array();
        $alphaLength = strlen($chars) - 1;
        for ($i = 0; $i < $keyLength; $i++) {
            $n = rand(0, $alphaLength);
            $pass[] = $chars[$n];
        }
        return implode($pass); //Turn the array into a string
    }

    /**
     * @param $date
     * @param string $format
     * @return bool
     *
     * Returns true if the supplied date is formatted as specified by $format
     */
    public static function validateDate($date, $format = 'Ymd') {

        $d = DateTime::createFromFormat($format, $date);
        return $d && $d->format($format) == $date && $date != '';
    }

    /**
     * @param $date
     * @param string $output
     * @param string $input
     * @return string
     *
     *      Returns a formatted date string using the given input and output parameters
     * Supports all PHP DateTime formats
     */
    public static function formatDate($date, $output = 'Y-m-d H:i:s', $input = 'Ymd') {
        $d = DateTime::createFromFormat($input, $date);
        return $d->format($output);
    }

    /**
     * @param $date
     * @param string $output
     * @param string $input
     * @return bool|string
     *
     * Returns a formatted date string of last Monday (or today if it is currently Monday)
     */
    public static function lastMonday($date, $output = 'Ymd', $input = 'Ymd') {
        // Returns a Ymd string of the last monday from the supplied date
        if ($input != 'Ymd')
            $date = View::formatDate($date, $output, $input);
        $date = strtotime($date);

        if (date('N', $date) == 1)
            $rv = date($output, $date);
        else
            $rv = date($output, strtotime(
                'last monday',
                $date
            ));

        return $rv;
    }

    /**
     * @param $val
     * @param $default
     * @return mixed
     *
     * Returns $default instead of $val if $val is:
     *      Null
     *      False
     *      0
     *      An empty string
     *      An empty array
     *
     * See http://php.net/manual/en/function.empty.php for all possible matching values
     */
    public static function defaultVal(&$val, $default) {
        if (empty($val))
            $val = $default;
        return $val;
    }

    /**
     * @throws RouteDeprecated
     *
     * Point a route or page at this function to return a Deprecation Notice
     */
    public function deprecated() {
        throw new RouteDeprecated($this->route);
    }

    /**
     * @throws RouteNotImplemented
     *
     * Point a route or page at this function to return a Not Implemented Notice
     */
    public function notImplemented() {
        throw new RouteNotImplemented($this->route);
    }

    public function errorHandler($f) {
        // Use this function to wrap view functions and provide error handling.

        $uID = $this->uID;
        $endPoint = get_class($this);
        try {
            $response = $f();
        }
        catch (ModelError $e){
            exception_handler($e);
            errorLog(array(
                'error'    => $e->name,
                'errorMsg' => $e->getMessage(),
                'endPoint' => $endPoint,
                'request'  => array(
                    'route' => $this->route,
                    $uID    => $this->ID,
                    'data'  => $this->DATA
                )
            ));
            new ErrorResponse($e);
        }
        catch (QueryError $e){
            exception_handler($e);
            errorLog(array(
                'error' => $e->name,
                'errorMsg' => $e->getMessage(),
                'endPoint' => $endPoint,
                'request'  => array(
                    'route' => $this->route,
                    $uID    => $this->ID,
                    'data'  => $this->DATA
                )
            ));
            new ErrorResponse($e);
        }
        catch (BaseError $e){
            new ErrorResponse($e);
        }


        if ($this->dbg)
            apiLog(array(
                'endPoint' => $endPoint,
                'request'  => array(
                    'route' => $this->route,
                    $uID     => $this->ID,
                    'data'   => $this->DATA
                ),
                'response' => array(
                    'success' => true,
                    'data'    => (count($response) > 2 ? count($response) . " objects" : $response)
                )
            ));


        if ($this->method == 'POST') {
            // Send success response if request completes without errors
            new SuccessResponse(View::prepResponseData($response));
        }
        else {
            if (!$response)
                $response = '';
            echo $response;
        }
    }
}

class APIView extends View {
    public $checkEndpoint = true;
    public $checkAction = true;

    public function checkMethod($method) {
        if ($method !== 'POST')
            throw new InvalidMethod("All API calls must be POST requests!");
        parent::checkMethod($method);
    }

    public function routeRequest($route) {
        if ($this->tracking === null)
            throw new MissingData("Tracking number is required");

        if ($this->endpoint === null && $this->checkEndpoint)
            throw new MissingData("Endpoint is required");

        if ($this->action === null && $this->checkAction)
            throw new MissingData("Action is required");

        return parent::routeRequest($route);
    }
}
