<?php

class ErrorCodes {
	const GENERIC_ERROR = 1;
	const MODEL_ERROR = 2;
	const VIEW_ERROR = 3;
}


/*
 * Base Exceptions
 */
class BaseError extends Exception {
	public $name;
	public $err;
	public $code = ErrorCodes::GENERIC_ERROR;
	public $data;

	public function __construct($err, $data = array()) {
		parent::__construct($err, $this->code, null);
		$this->err = $err;
		$this->data = $data;
	}
}
class TestError extends BaseError {
	public $name = 'Test';
}
class CronError extends BaseError {
	public $name = 'Cron Error';
}

// Authentication
class AuthFailed extends BaseError {
	public $name = 'Authentication Failed';
}
class AuthDenied extends BaseError {
	public $name = 'Permission Denied';
}
class LoginFailed extends BaseError {
    public $name = 'Login Failed';
}
class APIFailed extends BaseError {
    public $name = 'API Connection or Login Failed';
}
class ActivationFailed extends BaseError {
    public $name = 'Activation Failed';
}
class EmailFailure extends BaseError {
	public $name = 'Failure Sending Email';
}
class LoyaltyFailed extends BaseError {
    public $name = 'Loyalty Error';
}
class GiftCardFailed extends BaseError {
    public $name = 'Gift Card Error';
}

function extractErrorDetails(Exception $err) {
	$rv = array(
		'type' => 'Uncaught Exception',
		'msg' => $err->getMessage(),
		'file' => $err->getFile(),
		'line' => $err->getLine(),
		'trace' => print_r($err->getTrace(), 1)
	);

	if ($err instanceof BaseError) {
		$rv['type'] = $err->name;
	}

	if ($err instanceof QueryError) {
		if (!($err instanceof QueryNoLocDB))
			$rv['db'] = $err->query->dbName();

		$rv['query'] = $err->query->queryStr;
	}

	return $rv;
}

function format_error( $errno, $errstr, $errfile, $errline ) {

	$trace = print_r( debug_backtrace( false ), true );

	$content = "
  <table>
  <tr>
    <td><h3>Error</h3></td>
    <td><pre>$errstr</pre></td>
  </tr>
  <tr>
    <td><h3>Errno</h3></td>
    <td><pre>$errno</pre></td>
  </tr>
  <tr>
    <td><h3>File</h3></td>
    <td>$errfile</td>
  </tr>
  <tr>
    <td><h3>Line</h3></td>
    <td>$errline</td>
  </tr>
  <tr>
    <td><h3>Trace</h3></td>
    <td><pre>$trace</pre></td>
  </tr>
  </table>";
	return $content;

}

/*
 * Error Handlers
 */
function exception_handler(Exception $e) {
	global $Conf;
	$env = ($Conf->dbg ? "Dev - " : "Prod - ");

	$err = extractErrorDetails($e);

	$msg = "<div>
		<div><h3>Error:</h3><p>{$err['msg']}</p></div>
		<div><h3>File:</h3><p>{$err['file']}</p></div>
		<div><h3>Line:</h3<p>{$err['line']}</p></div>
		<div><h3>LocationID:</h3><p>{$_SESSION['locationID']}</p></div>
		";

	$subj = $err['type'];

	if ($e instanceof QueryError) {
		$msg .= "
			<li>Database: {$err['db']}</li>
			<li>Query: {$err['query']}</li>";
	}

	$msg .= "<div><h3>Stack Trace:</h3><pre>{$err['trace']}</pre></div></div>";

	$headers = 'MIME-Version: 1.0' . "\r\n";
	$headers .= 'From: debug@api.ordercounter.com' . "\r\n";
	$headers .= 'Content-type: text/html; charset=iso-8859-1' . "\r\n";

	mail(DEBUG_EMAIL, $env.$subj, $msg, $headers);

}

function error_handler($errno, $errstr, $errfile, $errline, $errarr) {

	global $Conf;
	$env = ($Conf->dbg ? "Dev - " : "Prod - ");

	$msg = format_error($errno,$errstr,$errfile,$errline);

	$headers = 'MIME-Version: 1.0' . "\r\n";
	$headers .= 'From: debug@api.ordercounter.com' . "\r\n";
	$headers .= 'Content-type: text/html; charset=iso-8859-1' . "\r\n";

	mail(DEBUG_EMAIL, $env.'Error: ' . basename($errfile) . " Line:{$errline}", $msg, $headers);

}

function fatal_error_handler() {

	$error = error_get_last();

	global $Conf;
	$env = ($Conf->dbg ? "Dev - " : "Prod - ");

	// Catch fatal errors, warnings and parse errors
	if ($error !== NULL && $error['type'] <= 2) {
		$errType = ($error['type'] === 1 ? 'Fatal Error' : 'Warning');

		$errno   = $error["type"];
		$errfile = $error["file"];
		$errline = $error["line"];
		$errstr  = $error["message"];

		$msg = format_error($errno,$errstr,$errfile,$errline);

		$headers = 'MIME-Version: 1.0' . "\r\n";
		$headers .= 'From: debug@api.ordercounter.com' . "\r\n";
		$headers .= 'Content-type: text/html; charset=iso-8859-1' . "\r\n";

		mail(DEBUG_EMAIL, $env ."{$errno}: " . basename($error['file']) . " on line {$error['line']}", $msg, $headers);

	}
	exit();
}

set_exception_handler('exception_handler');
set_error_handler('error_handler', E_ALL & ~E_NOTICE & ~E_WARNING);
register_shutdown_function('fatal_error_handler');
