<?php

/*
 * Model Exceptions
 */
class ModelError extends BaseError {
	public $name = 'Model Error';
	public $code = ErrorCodes::MODEL_ERROR;
}

class ModelNotFound extends ModelError {
	public $name = 'Model Not Found';
}


/**
 * Class DBModel
 *
 * The base Model class allows entries in the database to be mapped to PHP objects
 */
class DBModel {
	protected $table; // DB table
	protected $encrypted = array(); // Encrypted fields
	protected $hidden = array(); // Hidden fields which aren't saved to DB
    protected $dbg;
    protected $conn;

	// Exclude fields from export
	protected $exclude = array();

	function __construct($dbg = false, $conn = null) {
		$this->dbg = $dbg;

        if ($conn !== null)
            $this->conn = $conn;
        else {
            global $DBConnection;
            $this->conn = $DBConnection;
        }
	}

	public function __toString() {
		return print_r($this->export(), 1);
	}

	/**
	 * @param $filter
	 * @param $dbg
	 * @return Model|null
	 */
	public static function find($filter, $dbg = false, $conn = null) {
		/**
		 * @var $model Model
		 */
		$class = get_called_class();
		$model = new $class($dbg);
		$keys = implode(', ', $model->prepKeys());
		$find = $model->prepFilter($filter);

		if (isset($filter['orderBy']))
			$orderBy = " ORDER BY {$filter['orderBy']}";
		else
			$orderBy = "";

		$limit = " LIMIT 1";

		$qStr = "SELECT {$keys} FROM {$model->table} WHERE {$find}{$orderBy}{$limit}";
		$q = $model->query($qStr, $dbg, $conn);

		if ($q->check()) {
			$data = $q->arr();
			$model->log($data);
			return $model->load($data);
		}
		else
			return null;
	}

	public static function findAND($args, $dbg = false) {
		return static::find(array(
			'filter' => array(
				'type' => 'AND',
				'args' => $args
			)
		), $dbg);
	}

	/**
	 * @param $filter
	 * @param $dbg
	 * @return Model[]|null
	 */
	public static function findAll($filter, $dbg = false, $conn = null) {
		/*
		 * Returns models which match the given filter
		 * where $filter is an array of the following form:
		 *
		 * $filter = array(
		 *      'type' => '' // Filter operator: 'AND', 'OR', 'IN', etc... - required
		 *      'filter' => array(
		 *          array(
		 *              'field' => ''// Name of column in DB - required
		 *              'val' => '' // Value to match against - required
		 *              'op' => '' // Comparison operator: '=', '!=', '>', '>=', etc... - optional, defaults to '='
		 *              'raw' => '' // Flag to toggle escaping of field and val args - optional, defaults to false
		 *          ),
		 *          array( // as many filter operands as required for query
		 *              ...
		 *          )
		 *      )
		 * )
		 *
		 */
		/**
		 * @var $model Model
		 */

		$class = get_called_class();
		$model = new $class($dbg);
		$keys = implode(', ', $model->prepKeys());
		$find = $model->prepFilter($filter);

		if (isset($filter['orderBy']))
			$orderBy = " ORDER BY {$filter['orderBy']}";
		else
			$orderBy = "";

		if (isset($filter['limit']))
			$limit = " LIMIT {$filter['limit']}";
		else
			$limit = "";

		$qStr = "SELECT {$keys} FROM {$model->table} WHERE {$find}{$orderBy}{$limit}";

		$q = $model->query($qStr, false, $conn);

		if ($q->check()) {
			// Returns an array of $class objects
			return $q->all(function ($model) use ($class, $dbg) {
				return (new $class())->load($model);
			});
		}
		else
			return array();
	}

	public static function findAllAND($args, $dbg = false) {
		return static::findAll(array(
			'filter' => array(
				'type' => 'AND',
				'args' => $args
			)
		), $dbg);
	}

	public static function all($filter = array(), $dbg = false) {
		/**
		 * @var $model Model
		 */
		$class = get_called_class();
		$model = new $class($dbg);
		$keys = implode(', ', $model->prepKeys());
		$find = '1 = 1';

		if (isset($filter['orderBy']))
			$orderBy = " ORDER BY {$filter['orderBy']}";
		else
			$orderBy = "";

		if (isset($filter['limit']))
			$limit = " LIMIT {$filter['limit']}";
		else
			$limit = "";

		$qStr = "SELECT {$keys} FROM {$model->table} WHERE {$find}{$orderBy}{$limit}";

		$q = $model->query($qStr);

		if ($q->check()) {
			// Returns an array of $class objects
			return $q->all(function ($model) use ($class, $dbg) {
				return (new $class($dbg))->load($model);
			});
		}
		else
			return array();
	}

	public static function in($field, $data, $filter = array(), $dbg = false) {
		/*
		 * Returns models which match the given filter
		 * where $filter is an array of the following form:
		 *
		 * $filter = array(
		 *      'type' => '' // Filter operator: 'AND', 'OR', 'IN', etc... - required
		 *      'filter' => array(
		 *          array(
		 *              'field' => ''// Name of column in DB - required
		 *              'val' => '' // Value to match against - required
		 *              'op' => '' // Comparison operator: '=', '!=', '>', '>=', etc... - optional, defaults to '='
		 *              'raw' => '' // Flag to toggle escaping of field and val args - optional, defaults to false
		 *          ),
		 *          array( // as many filter operands as required for query
		 *              ...
		 *          )
		 *      )
		 * )
		 *
		 */

		/**
		 * @var $model Model
		 */
		$class = get_called_class();
		$model = new $class($dbg);
		$keys = implode(', ', $model->prepKeys());

		$imploded = implode(', ', $data);
		$find = "{$field} IN ({$imploded})";

		if (isset($filter['orderBy']))
			$orderBy = " ORDER BY {$filter['orderBy']}";
		else
			$orderBy = "";

		if (isset($filter['limit']))
			$limit = " LIMIT {$filter['limit']}";
		else
			$limit = "";

		$qStr = "SELECT {$keys} FROM {$model->table} WHERE {$find}{$orderBy}{$limit}";

		$q = $model->query($qStr);

		if ($q->check()) {
			// Returns an array of $class objects
			return $q->all(function ($model) use ($class, $dbg) {
				return (new $class())->load($model);
			});
		}
		else
			return null;
	}

	/**
	 * @param $op
	 * @param $args
	 * @param bool $dbg
	 * @return array|Model[]|null
	 */
	public static function where($args, $op = null, $dbg = false) {

		$parsedArgs = array();
		foreach ($args as $field => $val)
			$parsedArgs[] = array(
				'field' => $field,
				'val' => $val
			);

		if ($op)
			$models = self::findAll(array(
				'filter' => array(
					'type' => $op,
					'args' => $parsedArgs
				)
			), $dbg);
		else
			$models = self::findAll($parsedArgs, $dbg);

		return ($models !== null ? $models : array());
	}

	/**
	 * @param $field
	 * @param $models
	 * @return array
	 *
	 * Usage: Given an array of models, extract $field from each and return as an array
	 */
	public static function extract($field, $models) {
		$rv = array();
		if ($models)
			foreach ($models as $Model)
				$rv[] = $Model->$field;
		return $rv;
	}

	protected function prepFilter($filter) {
		if (isset($filter['field']))
			return $this->parseFilter($filter);
		else if (isset($filter['filter'])) {
			return $this->prepCondition($filter['filter']);
		}
		else {
			$tmp = '';
			foreach ($filter as $sub) {
				$tmp .= $this->prepFilter($sub);
			}
			return $tmp;
		}
	}

	protected function prepCondition($filter) {
		$tmp = array();
		foreach ($filter['args'] as $v)
			$tmp[] = $this->prepFilter($v);
		$rv = implode(" {$filter['type']} ", $tmp);
		$rv = '(' . $rv . ')';
		return $rv;
	}

	protected function parseFilter($filter) {
		if (!isset($filter['op']))
			$filter['op'] = '=';
		if (isset($filter['raw']) && $filter['raw'])
			return "{$filter['field']} {$filter['op']} {$filter['val']}";
		else
			return "`{$filter['field']}` {$filter['op']} '{$filter['val']}'";
	}

	/**
	 * @param $data
	 * @return $this
	 *
	 * Load data into this model
	 */
	public function load($data) {
		$fields = array_keys($this->fields());
		if (count($data))
			foreach ($data as $k => $v) {
				if (in_array($k, $fields))
					$this->$k = $v;
			}
		return $this;
	}

	/**
	 * @return array
	 *
	 * Returns an array of this model's fields and their corresponding values
	 */
	public function fields() {
		$rv = array();
		// Grab all props of this model
		// Filter out private(*) and protected(get_class) props and DB key
		// Leave only fields that are in the DB
		foreach ((array)$this as $k => $v)
			if (strpos($k, '*') !== 1 && strpos($k, get_class($this)) !== 1 && (!$this->hidden || !in_array($k, $this->hidden)))
				$rv[$k] = $v;
		return $rv;
	}

	public function prepKeys() {
		$keys = array();

		foreach ($this->fields() as $k => $v) {
			if (in_array($k, $this->encrypted))
				$keys[] = "AES_DECRYPT(`{$k}`, '". Conf::secretKey ."') AS `{$k}`";
			else
				$keys[] = "`{$k}`";
		}
		return $keys;
	}

	public function prepVals() {
		$vals = array();

		foreach ($this->fields() as $k => $v) {
			if (in_array($k, $this->encrypted))
				$vals[] = "AES_ENCRYPT(\"{$v}\", '". Conf::secretKey ."')";
			else
				$vals[] = "\"" . mysql_real_escape_string($v) . "\"";
		}
		return $vals;
	}

	/**
	 * @return array
	 *
	 * This function is used to export the model data into a format suitable for consumption by API
	 * Return anything needed by API interactions that isn't sensitive (passwords, etc)
	 */
	public function export() {
		$rv = $this->fields();

		if ($this->exclude)
			foreach($this->exclude as $item)
				unset($rv[$item]);

		return $rv;
	}

	/**
	 * @param string $format
	 * @param null $time
	 * @return bool|string
	 *
	 * Returns a formatted date string of the current time
	 */
	public static function currentDate($format = 'Y-m-d H:i:s', $time = null) {
		if ($time === null)
			$time = time();
		return date($format, $time);
	}

	/**
	 * @return int
	 *
	 * Returns a timestamp of the current time
	 */
	public static function timestamp() {
		return time();
	}

	public static function toTimestamp($date) {
		return strtotime($date);
	}

	protected function log($str) {
		if ($this->dbg)
			miscLog($str);
	}

	protected function query($str, $dbg = false, $conn = null) {
        if ($conn !== null) {
            $this->conn = $conn;
        }

		return new Query($str, $this->dbg || $dbg, $this->conn);
	}
}

/**
 * Class Model
 *
 * The Model class provides functionality for entries database with a Primary Key
 *
 */
class Model extends DBModel {
	protected $key;
	protected $ID;

	public function create($data = array(), $usingKey = false, $dbg = false, $conn = null) {
		$this->load($data);
		$prep = $this->prepInsert($usingKey);

		$query = $this->query("INSERT INTO {$this->table} ({$prep['keys']}) VALUES ({$prep['vals']})", $dbg, $conn);
		if ($query->updated()) {
			if (!$usingKey) {
				$this->ID = $query->newID();
				$this->{$this->key} = $this->ID;
			}
			$this->log("ID: ".$this->ID);
		}
		else
			throw new Exception();
		return $this;
	}

	public function read($ID) {
		$keys = implode(', ', $this->prepKeys());
		$q = $this->query("SELECT {$keys} FROM {$this->table} WHERE {$this->key} = '{$ID}'");
		if ($q->check()) {
			$this->ID = $this->{$this->key} = $ID;
			$model = $q->arr();

			$this->log($model);
			$this->load($model);
			return $this;
		}
		else {
			$class = get_class($this);
			throw new ModelNotFound("Invalid {$class} ID: [{$ID}]");
		}
	}

	public function update($data = array()) {
		$this->load($data);

		$prepKeys = array_keys($this->fields());
		$prepVals = $this->prepVals();
		$fields = array();

		foreach ($prepKeys as $index => $key)
			if ($key != $this->key) {
				$fields[] = "`{$key}` = {$prepVals[$index]}";
			}
		$fields = implode(', ', $fields);

		$this->query("UPDATE {$this->table} SET {$fields} WHERE {$this->key} = '{$this->ID}'");
		return $this;
	}

	public function delete() {
		$this->query("DELETE FROM {$this->table} WHERE {$this->key} = '{$this->ID}'");
		return $this;
	}

	public function load($data) {
		foreach($data as $k => $v)
			if ($this->key == $k)
				$this->{$this->key} = $this->ID = $v;
		return parent::load($data);
	}

	public function export() {
		$rv = parent::export();
		$rv[$this->key] = (int)$rv[$this->key];
		return $rv;
	}

	public function prepInsert($usingKey = false) {
		$keys = array();
		$vals = array();

		foreach ($this->fields() as $k => $v) {
			if (($v != '' || $v !== null) && ($usingKey || $k != $this->key)) {
				$keys[] = "`{$k}`";
				if (in_array($k, $this->encrypted))
					$vals[] = "AES_ENCRYPT(\"{$v}\", '". Conf::secretKey ."')";
				else
					$vals[] = "\"" . mysql_real_escape_string($v) . "\"";
			}
		}
		return array(
			'keys' => implode(', ', $keys),
			'vals' => implode(', ', $vals)
		);
	}
}


/**
 * Class Map
 *
 * The Map class is used for models that don't have a primary key and which map relationships (ID1, ID2, ID3)
 */
class Map extends DBModel {
	public static function create($data, $dbg = false) {
		/**
		 * @var $model Model
		 */
		$class = get_called_class();
		$model = new $class($dbg);

		$model->log($data);
		$model->load($data);

		$prep = self::prepInsert($model);

		$model->query("INSERT INTO {$model->table} ({$prep['keys']}) VALUES ({$prep['vals']})");

		return $model;
	}

	public static function update($filter, $data, $dbg = false) {
		/**
		 * @var $model Model
		 */
		$class = get_called_class();
		$model = new $class($dbg);

		$find = $model->prepFilter($filter);

		$model->load($data);
		$prepKeys = array_keys($model->fields());
		$prepVals = $model->prepVals();

		$fields = array();

		foreach ($prepKeys as $index => $key)
			if ($prepVals[$index] !== '')
				$fields[] = "`{$key}` = {$prepVals[$index]}";

		$fields = implode(', ', $fields);
		$model->query("UPDATE {$model->table} SET {$fields} WHERE {$find}");

		return $model;
	}

	public static function delete($filter, $dbg = false) {
		/**
		 * @var $model Model
		 */
		$class = get_called_class();
		$model = new $class($dbg);

		$find = $model->prepFilter($filter);

		$model->query("DELETE FROM {$model->table} WHERE {$find}");
		return null;
	}

	/**
	 * @param Map $map
	 * @return array
	 */
	public static function prepInsert($map) {
		$keys = array();
		$vals = array();

		foreach ($map->fields() as $k => $v) {
			$keys[] = "`{$k}`";
			if (in_array($k, $map->encrypted))
				$vals[] = "AES_ENCRYPT(\"{$v}\", '". Conf::secretKey ."')";
			else
				$vals[] = "\"" . mysql_real_escape_string($v) . "\"";
		}
		return array(
			'keys' => implode(', ', $keys),
			'vals' => implode(', ', $vals)
		);
	}
}
